"""
Normalized sum of squared errors (nSSE)
"""

import numpy as np
import pandas as pd


def save_table(num_c, num_f, err_arr, path):
    df_list = []
    for idx in range(num_c):
        df_list.append([idx, np.round(err_arr[idx], 2), num_f[idx]])

    df = pd.DataFrame(df_list, columns=['Class', 'nSSE', 'N'])
    df.to_csv(path, index=False)


def print_table(num_c, num_f, err_arr, title):
    avg_err = np.sum(err_arr) / num_c

    table = "|{:^30}|\n".format(title)
    table += "|{:^9}|{:^10}|{:^9}|\n".format('Class', 'nSSE', 'N')
    table += "-" * 32 + "\n"
    for idx in range(num_c):
        table += "|{:^9}|{:^10}|{:^9}|\n".format(idx, np.round(err_arr[idx], 2), num_f[idx])
    table += "-" * 32 + "\n"
    table += "Average nSSE: {}".format(np.round(avg_err, 2))

    print(table)


def find_nearest_idx(anchor, centroids):
    dist = np.sum(np.square(centroids - anchor), axis=1)
    idx = np.argmin(dist)
    return idx


def cal_norm_factor(anchor, centroids):
    idx = find_nearest_idx(anchor, centroids)
    norm_factor = np.sum(np.square(centroids[idx] - anchor))
    return norm_factor


def sse_per_class(feature, centroid):
    return np.sum(np.sum(np.power(centroid - feature, 2), axis=1), axis=0)


def nsse(
        feature,
        label,
        opt_print=False,
        opt_save=False,
        path=None,
):
    num_c = np.array(list(set(label))).shape[0]
    num_f = np.zeros(num_c)
    err_arr = np.zeros(num_c)
    centroids = np.zeros((num_c, feature.shape[-1]))

    for idx in range(num_c):
        f_idx = np.where(label == idx)[0]
        centroids[idx] = np.mean(feature[f_idx], axis=0)

    for idx in range(num_c):
        f_idx = np.where(label == idx)[0]
        norm_factor = cal_norm_factor(centroids[idx], np.delete(centroids, idx, 0))
        err_arr[idx] = sse_per_class(feature[f_idx], centroids[idx]) / norm_factor
        num_f[idx] = f_idx.shape[0]

    avg_err = np.sum(err_arr) / num_c

    if opt_print:
        print_table(num_c, num_f, err_arr, '`toolkits.cluster.nsse`')
    if opt_save:
        assert path is not None, "Please enter the save path."
        save_table(num_c, num_f, err_arr, path)

    return avg_err


def batch_nsse(
        feature,
        label,
        opt_print=False,
        opt_save=False,
        path=None,
):
    assert len(label.shape) >= 2, "The input form is not batch."

    num_batch = label.shape[0]
    num_c = np.array(list(set(label[0]))).shape[0]
    num_f = np.zeros((num_batch, num_c))
    err_arr = np.zeros((num_batch, num_c))
    centroids = np.zeros((num_batch, num_c, feature.shape[-1]))

    for i in range(num_batch):
        for idx in range(num_c):
            f_idx = np.where(label[i] == idx)[0]
            centroids[i, idx] = np.mean(feature[i, f_idx], axis=0)

        for idx in range(num_c):
            f_idx = np.where(label[i] == idx)[0]
            norm_factor = cal_norm_factor(centroids[i, idx], np.delete(centroids[i], idx, 0))
            err_arr[i, idx] = sse_per_class(feature[i, f_idx], centroids[i, idx]) / norm_factor
            num_f[i, idx] = f_idx.shape[0]

    num_f = np.mean(num_f, axis=0)
    err_arr = np.mean(err_arr, axis=0)
    avg_err = np.sum(err_arr) / num_c

    if opt_print:
        print_table(num_c, num_f, err_arr, '`toolkits.cluster.batch_nsse`')
    if opt_save:
        assert path is not None, "Please enter the save path."
        save_table(num_c, num_f, err_arr, path)

    return avg_err



"""
Simple print for predictions and true labels
"""

import numpy as np
import pandas as pd


def save_table(num_c, label_f, pred_arr, path):
    df_list = []
    for idx in range(num_c):
        f_idx = np.where(label_f == idx)[0]
        pred_list = list(np.array(pred_arr[f_idx], dtype=int))
        df_list.append([idx, len(pred_list), pred_list.count(idx), pred_list])

    df = pd.DataFrame(df_list, columns=['Class', 'N', 'Correct', 'Predict'])
    df.to_csv(path, index=False)


def print_table(num_c, label_f, pred_arr):
    table = "{:^9}|{:^9}|{:^9}|{:^9}\n".format('Class', 'N', 'Correct', 'Predict')
    table += "-" * 39 + "\n"
    for idx in range(num_c):
        f_idx = np.where(label_f == idx)[0]
        pred_list = list(np.array(pred_arr[f_idx], dtype=int))
        pred_str = ""
        for item in pred_list:
            pred_str += "{:^7}".format(str(item))
        table += "{:^9}|{:^9}|{:^9}|{}\n".format(idx, len(pred_list), pred_list.count(idx), pred_str)

    print(table)


def get_logit(feature, classifier, similarity):
    if similarity == 'euclidean':
        w = np.expand_dims(classifier, 0)
        q = np.expand_dims(feature, 1)
        logit = -np.sum(np.square(q - w), axis=2)
    else:
        assert False, "Not supported similarity"
    return logit


def pred_summary(
        feature,
        classifier,
        label_f,
        label_c,
        similarity='euclidean',
        opt_save=False,
        path=None,
):
    num_c = np.array(list(set(label_c))).shape[0]
    num_f = feature.shape[0]
    pred_arr = np.zeros(num_f)

    for idx in range(num_c):
        f_idx = np.where(label_f == idx)[0]
        logit = get_logit(feature[f_idx], classifier, similarity)
        pred_arr[f_idx] = np.argmax(logit, axis=1)

    print_table(num_c, label_f, pred_arr)

    if opt_save:
        assert path is not None, "Please enter the save path."
        save_table(num_c, label_f, pred_arr, path)



"""
The `toolkits.parse` module implements log parser.
"""

from .between_lines import between_lines, between_lines_on_file, between_lines_on_dir

__all__ = [
    "between_lines",
    "between_lines_on_file",
    "between_lines_on_dir",
]



"""
Visualize features and classifiers using t-SNE plot.
"""

import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import numpy as np


def make_colors(y_set):
    n_label = len(y_set)
    np.random.seed(42)
    colors = [[np.random.choice(range(256)) / 256,
               np.random.choice(range(256)) / 256,
               np.random.choice(range(256)) / 256]
              for _ in range(n_label)]
    return colors


def tsne(feature, classifier, label_f, label_c, save_path, perplexity=30, seed=42):
    """ t-SNE plot for features and classifiers
    :param feature: Feature vectors(n_data, dim)
    :param classifier : Classifier vectors (n_classes, dim)
    :param label_f : Label array of input feature vectors
    :param label_c:  Label array of input classifier vectors
    :param save_path: Save path of t-SNE result
    :param perplexity: Perplexity
    :param seed: Random seed
    """
    model = TSNE(random_state=seed, perplexity=perplexity)

    x = np.concatenate([feature, classifier])
    y = np.concatenate([label_f, label_c])
    y_set = list(set(y))
    colors = make_colors(y_set)

    out = model.fit_transform(x)

    for i in range(feature.shape[0]):
        c_idx = y_set.index(y[i])
        plt.scatter(out[i, 0], out[i, 1], color=colors[c_idx], s=10)

    for i in range(classifier.shape[0]):
        idx = feature.shape[0] + i
        plt.scatter(out[idx, 0], out[idx, 1], color='blue', s=70, marker='X')
        plt.text(out[idx, 0], out[idx, 1], str(int(y[idx])), fontdict={'size': 13, 'color': 'white', 'fontweight':1000})
        plt.text(out[idx, 0], out[idx, 1], str(int(y[idx])), fontdict={'size': 13})

    size = 2.5 * np.std(out)
    plt.xlim(-size, size)
    plt.ylim(-size, size)

    plt.show()
    plt.savefig(save_path)
    plt.close()



"""
File management tools.
"""

import pickle as pkl
import numpy as np


def save(feature, classifier, label_f, label_c, file_path):
    """Save vectors ans labels"""
    file_format = file_path.split('.')[-1]

    if file_format == 'pkl':
        data = {"feature": feature, "classifier": classifier,
                "label_f": label_f, "label_c": label_c}
        with open(file_path, "wb") as f:
            pkl.dump(data, f)
    elif file_format == 'npz':
        np.savez(file_path, feature=feature, classifier=classifier, label_f=label_f, label_c=label_c)
    else:
        assert False, "Not supported data format."


def load(file_path):
    """Load vectors ans labels from data file"""
    file_format = file_path.split('.')[-1]

    if file_format == 'pkl':
        with open(file_path, "rb") as f:
            data = pkl.load(f)
            feature = data['feature']
            classifier = data['classifier']
            label_f = data['label_f']
            label_c = data['label_c']
    elif file_format == 'npz':
        data = np.load(file_path)
        feature = data['feature']
        classifier = data['classifier']
        label_f = data['label_f']
        label_c = data['label_c']
    else:
        assert False, "Not supported data format."

    return feature, classifier, label_f, label_c



"""
The `toolkits.cluster` module implements cluster performance analysis techniques.
"""

from .sse import sse, batch_sse
from .nsse import nsse, batch_nsse
from .nearc import nearc

__all__ = [
    "sse",
    "batch_sse",
    "nsse",
    "batch_nsse",
    "nearc",
]


import os


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    libraries = []
    if os.name == "posix":
        libraries.append("m")

    config = Configuration("toolkits", parent_package, top_path)
    config.add_subpackage("cluster")
    config.add_subpackage("pprint")
    config.add_subpackage("utils")
    config.add_subpackage("viz")
    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(**configuration(top_path="").todict())



"""
Extract the log between the two input sentences
"""
import os


def _file2list(path):
    """ Get file content to list type
    :param path: file path
    :return: file content list
    """
    with open(path, 'r') as f:
        lines = f.read().splitlines()
    return lines


def _between_lines(lines, beg, end, single=False):
    """ Extract lines from `beg` to `end`
    :param lines: content list which is split for each newline
    :param beg: starting point of parsing
    :param end: end point of parsing
    :param single: whether there are multiple sections to exist
    :return: extracted lines
    """

    # Initialize variables
    start_idx = -1
    end_idx = -1
    indices = []

    # Parsing
    for idx, line in enumerate(lines):
        if line.startswith(beg):
            start_idx = idx
        if line.startswith(end) and start_idx != -1:
            end_idx = idx

        if start_idx != -1 and end_idx != -1:
            assert end_idx >= start_idx, "must be 'end_idx >= start_idx'"
            indices.append([start_idx, end_idx])

            start_idx = -1
            end_idx = -1

            if single:
                break

    return [lines[item[0]:item[1]] for item in indices]


def between_lines(input_str, beg, end, single=False):
    """ Extract lines from `beg` to `end`
    :param input_str: input string
    :param beg: starting point of parsing
    :param end: end point of parsing
    :param single: whether there are multiple sections to exist
    :return: extracted lines
    """
    lines = input_str.splitlines()
    return _between_lines(lines, beg, end, single)


def between_lines_on_file(file_path, beg, end, single=False):
    """ Extract lines from `beg` to `end` from the target file
    :param file_path: file path
    :param beg: starting point of parsing
    :param end: end point of parsing
    :param single: whether there are multiple sections to exist
    :return: extracted lines
    """
    lines = _file2list(file_path)
    return _between_lines(lines, beg, end, single)


def between_lines_on_dir(dir_path, beg, end, single=False, path_filter=None, make_file=False, result_file='result.txt'):
    """ Extract lines from `beg` to `end` from the target directory
    :param dir_path: directory path
    :param beg: starting point of parsing
    :param end: end point of parsing
    :param single: whether there are multiple sections to exist
    :param path_filter: search only files containing `path_filter`
    :param make_file: boolean flag for making result file
    :param result_file: result file name
    :return: dictionary (key: file path, value: extracted lines)
    """
    file_list = list()
    for (root, directories, files) in os.walk(dir_path):
        for target_file in files:
            file_list.append(os.path.join(root, target_file))

    if path_filter:
        file_list = [item for item in file_list if path_filter in item]

    file2lines = dict()

    for file_path in file_list:
        file2lines[file_path] = between_lines_on_file(file_path, beg, end, single)

    if make_file:
        with open(result_file, 'w') as f:
            for k, v in file2lines.items():
                f.write(f"<target file: {k}>\n")
                for item in v:
                    f.write("\n".join(item))
                f.write("\n")

    return file2lines



"""
Parameter manipulation tools.
"""

import math
import numpy as np
import torch


def freeze_selected_param(model, target):
    """ Freeze layers which is in 'target'
    notice) This feature has not yet been tested.
    :param model: pytorch model
    :param target: layer name list
    :return: processed model
    """
    for param in model.parameters():
        param.requires_grad = True

    for name, param in model.named_parameters():
        if name in target:
            param.requires_grad = False

    return model


def get_important_param_idx(model, ratio, inverse=False):
    """ Get indices of parameters which has large absolute value
    :param model: pytorch model
    :param ratio: ratio of important parameters
    :param inverse: if inverse=True, select unimportant weights
    :return: dictionary of the indices of selected parameters
    """
    param_dict = dict(model.named_parameters())

    # Remove not trainable parameter from parameter dictionary
    remove_list = [k for k, v in param_dict.items() if not v.requires_grad]
    for k in remove_list:
        param_dict.pop(k)

    # Pre-processing parameter data
    param_flatten_dict = {}
    param_len_dict = {}
    for k, v in param_dict.items():
        param_dict[k] = param_dict[k].data
        param_flatten_dict[k] = torch.reshape(param_dict[k].detach().data, [-1])
        param_len_dict[k] = param_flatten_dict[k].shape[0]

    # Select the important weight (large absolute value)
    num_trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    num_st = math.floor(num_trainable_params * ratio)  # The number of session trainable parameters
    all_weight = np.abs(np.array(sum([list(v.cpu().numpy()) for v in param_flatten_dict.values()], [])))
    if inverse:
        sel_weight_idx = all_weight.argsort()[:num_st]
    else:
        sel_weight_idx = all_weight.argsort()[::-1][:num_st]

    # Get indices of selected parameters
    count = 0
    param_idx_dict = {}
    for k, v in param_len_dict.items():
        param_flattened_idx_list = [w_idx - count for w_idx in sel_weight_idx
                                    if (w_idx >= count) & (w_idx < count + v)]
        param_idx_dict[k] = [np.unravel_index(flattened_idx, param_dict[k].shape)
                             for flattened_idx in param_flattened_idx_list]
        count += v

    return param_idx_dict




"""
top-N nearest interfering centroid
"""

import numpy as np


def print_table(result):
    num_c = result.shape[0]
    N = result.shape[1]

    table = "{:^9}".format("Class")
    for i in range(N):
        table += "|{:^7}".format("Near{}".format(i+1))
    table += "\n" + "-" * (N * 10)

    for idx in range(num_c):
        table += "\n{:^9}|".format(idx)
        for i in range(N):
            table += "{:^8}".format(int(result[idx, i]))

    print(table)


def get_nearest(N, idx, centroids):

    anchor = centroids[idx]
    dist = np.sum(np.square(centroids - anchor), axis=1)

    result = []
    for i in range(N):
        top_n_idx = np.argsort(dist)[i+1]     # exclude itself (exclude 0th result)
        result.append(top_n_idx)

    return result


def nearc(feature, label, N=5, opt_print=False):
    num_c = np.array(list(set(label))).shape[0]
    num_f = np.zeros(num_c)
    centroids = np.zeros((num_c, feature.shape[-1]))

    result = np.zeros((num_c, N))

    for idx in range(num_c):
        f_idx = np.where(label == idx)[0]
        centroids[idx] = np.mean(feature[f_idx], axis=0)

    for idx in range(num_c):
        f_idx = np.where(label == idx)[0]
        result[idx] = get_nearest(N, idx, centroids)
        num_f[idx] = f_idx.shape[0]

    if opt_print:
        print_table(result)

    return result


import os
from setuptools import setup, find_packages


def configuration(parent_package="", top_path=None):
    if os.path.exists("MANIFEST"):
        os.remove("MANIFEST")

    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)
    config.add_subpackage("toolkits")
    return config


def setup_package():
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()

    metadata = dict(
        name='dl-toolkits',
        version='1.1.1',
        long_description=long_description,
        long_description_content_type='text/markdown',
        description='Deep Learning Analysis Toolkits',
        author='Yuho Jeong',
        url='https://github.com/yuhodots/toolkits',
        install_requires=[],
        packages=find_packages(exclude=[]),
        keywords=['dl', 'toolkits'],
        python_requires='>=3',
    )
    metadata["configuration"] = configuration

    setup(**metadata)


if __name__ == "__main__":
    setup_package()
